import collections
import io
import itertools
import os
import operator
import pathlib
import typing
import random
from typing import List, Optional, Iterator, Iterable, Collection, Type, Callable

import rich.progress
import numpy

from .._log import ConsoleLogger
from .._utils import guess_sequences_format, MofNCompleteUnitColumn
from ..._meta import zopen

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files  # type: ignore

if typing.TYPE_CHECKING:
    from Bio.SeqRecord import SeqRecord
    from ...model import Gene
    from ...hmmer import HMM
    from ...crf import ClusterCRF


# --- Output files -------------------------------------------------------------


def make_output_directory(
    logger: ConsoleLogger, output_dir: pathlib.Path, outputs: List[str]
) -> None:
    # Make output directory
    logger.info("Using", "output folder", repr(str(output_dir)), level=1)
    os.makedirs(output_dir, exist_ok=True)

    # Check if any output file already exist
    for output in outputs:
        if os.path.isfile(os.path.join(output_dir, output)):
            logger.warn("Output folder contains files that will be overwritten")
            break


def write_genes_table(
    logger: ConsoleLogger,
    genes: List["Gene"],
    *,
    genome: pathlib.Path,
    output_dir: pathlib.Path,
) -> None:
    from ...model import GeneTable

    base, _ = os.path.splitext(os.path.basename(genome))
    pred_out = os.path.join(output_dir, f"{base}.genes.tsv")
    logger.info("Writing", "gene table to", repr(pred_out), level=1)
    with open(pred_out, "wb") as f:
        GeneTable.from_genes(genes).dump(f)


def write_feature_table(
    logger: ConsoleLogger,
    genes: List["Gene"],
    *,
    genome: pathlib.Path,
    output_dir: pathlib.Path,
) -> None:
    from ...model import FeatureTable

    base, _ = os.path.splitext(os.path.basename(genome))
    pred_out = os.path.join(output_dir, f"{base}.features.tsv")
    logger.info("Writing", "feature table to", repr(pred_out), level=1)
    with open(pred_out, "wb") as f:
        FeatureTable.from_genes(genes).dump(f)


def write_cluster_table(
    logger: ConsoleLogger,
    clusters: List["Cluster"],
    *,
    genome: pathlib.Path,
    output_dir: pathlib.Path,
):
    from ...model import ClusterTable

    base, _ = os.path.splitext(os.path.basename(genome))
    cluster_out = os.path.join(output_dir, f"{base}.clusters.tsv")
    logger.info("Writing", "cluster table to", repr(cluster_out), level=1)
    with open(cluster_out, "wb") as out:
        ClusterTable.from_clusters(clusters).dump(out)


def write_clusters(
    logger: ConsoleLogger,
    clusters: List["Cluster"],
    *,
    genome: pathlib.Path,
    output_dir: pathlib.Path,
    merge: bool = False,
) -> None:
    from Bio import SeqIO

    if merge:
        base, _ = os.path.splitext(os.path.basename(genome))
        gbk_out = os.path.join(output_dir, f"{base}.clusters.gbk")
        records = (cluster.to_seq_record() for cluster in clusters)
        logger.info("Writing", f"all clusters to", repr(str(gbk_out)), level=1)
        SeqIO.write(records, gbk_out, "genbank")
    else:
        for cluster in clusters:
            gbk_out = os.path.join(output_dir, f"{cluster.id}.gbk")
            logger.info(
                "Writing",
                f"cluster [bold blue]{cluster.id}[/] to",
                repr(str(gbk_out)),
                level=1,
            )
            SeqIO.write(cluster.to_seq_record(), gbk_out, "genbank")


def write_sideload_json(
    logger: ConsoleLogger,
    clusters: List["Cluster"],
) -> None:
    raise NotImplementedError


# --- Load input (sequences) ---------------------------------------------------


def load_sequences(
    logger: ConsoleLogger,
    genome: pathlib.Path,
    *,
    format: Optional[str],
) -> Iterator["SeqRecord"]:
    from Bio import SeqIO

    # guess format or use the one given in CLI
    if format is not None:
        format: Optional[str] = format.lower()
        logger.info("Using", "user-provided sequence format", repr(format), level=2)
    else:
        logger.info("Detecting", "sequence format from file contents", level=2)
        format = guess_sequences_format(genome)
        if format is None:
            raise RuntimeError(f"Failed to detect format of {genome!r}")
        logger.success("Detected", "format of input as", repr(format), level=2)

    # load sequences from given file
    try:
        with logger.progress(download=True) as progress:
            n = 0
            logger.info("Loading", "sequences from genomic file", repr(genome), level=1)
            with progress.open(genome, "rb", description="Loading input") as file:
                with zopen(file) as decompressed:
                    reader = io.TextIOWrapper(decompressed)
                    for record in SeqIO.parse(reader, format):  # type: ignore
                        yield record
                        n += 1
    except ValueError as err:
        raise RuntimeError(f"Failed to load sequences: {err}") from err
    else:
        logger.success(f"Loaded {n} sequences from {str(genome)!r}", level=1)


# --- Load input (tables) ------------------------------------------------------


def load_genes(
    logger: ConsoleLogger,
    table_path: pathlib.Path,
) -> Iterator["Gene"]:
    from ...model import GeneTable

    # load gene table
    with logger.progress(download=True) as progress:
        logger.info("Loading", "genes table from file", repr(str(table_path)))
        with progress.open(table_path, "rb", description="Loading genes") as file:
            with zopen(file) as decompressed:
                table = GeneTable.load(decompressed)

    # count genes and yield gene objects
    with logger.progress() as progress:
        n_genes = len(set(table.protein_id))
        unit = "gene" if n_genes == 1 else "genes"
        task = progress.add_task("Building genes", unit=unit, total=n_genes)
        yield from progress.track(table.to_genes(), task_id=task)


def load_features(
    logger,
    table_paths: Iterable[pathlib.Path],
) -> "FeatureTable":
    from ...model import FeatureTable

    features = FeatureTable()
    with logger.progress(download=True) as progress:
        for filename in progress.track(table_paths, description="Loading features"):
            logger.info("Loading", "features table from file", repr(str(filename)))
            with progress.open(filename, "rb", description="Reading") as file:
                with zopen(file) as decompressed:
                    features += FeatureTable.load(decompressed)

    logger.success("Loaded", "a total of", len(features), "features", level=1)
    return features


def annotate_genes(
    logger: ConsoleLogger, genes: List["Gene"], features: "FeatureTable"
) -> List["Gene"]:
    from ...model import Domain

    # index genes by protein_id
    gene_index = {gene.protein.id: gene for gene in genes}
    if len(gene_index) < len(genes):
        raise ValueError("Duplicate gene names in input genes")

    # add domains from the feature table
    with logger.progress() as progress:
        unit = "row" if len(features) == 1 else "rows"
        task = progress.add_task("Annotating", unit=unit, total=len(features))
        for i in progress.track(range(len(features)), task_id=task):
            # get gene by ID and check consistency
            length = features.end[i] - features.start[i]
            gene = gene_index[features.protein_id[i]]
            if gene.source.id != features.sequence_id[i]:
                raise ValueError(
                    f"Mismatched source sequence for {features.protein_id[i]!r}: {gene.source.id!r} != {features.sequence_id[i]!r}"
                )
            elif gene.end - gene.start != length:
                raise ValueError(
                    f"Mismatched gene length for {row.protein_id!r}: {gene.end - gene.start!r} != {length!r}"
                )
            elif gene.start != features.start[i]:
                raise ValueError(
                    f"Mismatched gene start for {row.protein_id!r}: {gene.start!r} != {features.start[i]!r}"
                )
            elif gene.end != features.end[i]:
                raise ValueError(
                    f"Mismatched gene end for {row.protein_id!r}: {gene.end!r} != {features.end[i]!r}"
                )
            elif gene.strand.sign != features.strand[i]:
                raise ValueError(
                    f"Mismatched gene strand for {row.protein_id!r}: {gene.strand.sign!r} != {features.strand[i]!r}"
                )
            # add the row domain to the gene
            domain = Domain(
                name=features.domain[i],
                start=features.domain_start[i],
                end=features.domain_end[i],
                hmm=features.hmm[i],
                i_evalue=features.i_evalue[i],
                pvalue=features.pvalue[i],
                probability=features.cluster_probability[i],
            )
            gene.protein.domains.append(domain)

    # return
    return list(gene_index.values())


def assign_sources(
    logger: ConsoleLogger,
    sequences: Iterable["SeqRecord"],
    genes: List["Gene"],
    *,
    genome: pathlib.Path,
) -> Iterable["Gene"]:
    from ...model import Protein, Strand

    logger.info("Extracting", "known sequences from gene list", level=2)
    known_sequences = {gene.source.id for gene in genes}

    logger.info("Building", "index of source sequences", level=2)
    sequence_index = {
        record.id: record for record in sequences if record.id in known_sequences
    }

    try:
        logger.info("Assigning", "source sequences to gene objects", level=2)
        for gene in genes:
            gene = gene.with_source(sequence_index[gene.source.id])
            gene_seq = gene.source.seq[gene.start - 1 : gene.end]
            if gene.strand == Strand.Reverse:
                gene_seq = gene_seq.reverse_complement()
            gene = gene.with_protein(gene.protein.with_seq(gene_seq.translate()))
            yield gene
    except KeyError as err:
        raise RuntimeError("Sequence {!r} not found in {!r}", gene.source.id, str(genome)) from err


def load_clusters(
    logger: ConsoleLogger,
    clusters: pathlib.Path,
) -> "ClusterTable":
    from ...model import ClusterTable

    logger.info("Loading", "clusters table from file", repr(str(clusters)))
    with logger.progress(download=True) as progress:
        with progress.open(clusters, "rb", description="Loading clusters") as file:
            with zopen(file) as decompressed:
                return ClusterTable.load(decompressed)  # type: ignore


def label_genes(
    logger: ConsoleLogger, genes: List["Gene"], clusters: "ClusterTable"
) -> List["Gene"]:

    cluster_by_seq = collections.defaultdict(list)
    for i in range(len(clusters)):
        cluster_by_seq[clusters.sequence_id[i]].append(
            (clusters.start[i], clusters.end[i])
        )

    with logger.progress() as progress:
        gene_count = len(genes)
        unit = "gene" if gene_count == 1 else "genes"
        task = progress.add_task(
            "Labelling genes", total=gene_count, unit=unit, precision=""
        )

        logger.info("Labelling", "genes belonging to clusters")
        labelled_genes = []
        for seq_id, seq_genes in itertools.groupby(
            genes, key=operator.attrgetter("source.id")
        ):
            for gene in seq_genes:
                if any(
                    cluster_start <= gene.end and gene.start <= cluster_end
                    for (cluster_start, cluster_end) in cluster_by_seq[seq_id]
                ):
                    gene = gene.with_probability(1)
                else:
                    gene = gene.with_probability(0)
                labelled_genes.append(gene)
                progress.update(task_id=task, advance=1)

    return labelled_genes


# --- Extract genes ------------------------------------------------------------


def extract_genes(
    logger,
    sequences: List["SeqRecord"],
    *,
    cds_feature: Optional[str],
    locus_tag: Optional[str],
    mask: bool,
    jobs: int,
) -> Iterable["Gene"]:
    from ...orf import PyrodigalFinder, CDSFinder

    logger.info("Extracting", "genes from input sequences", level=1)
    if cds_feature is None:
        logger.info("Using", "Pyrodigal in metagenomic mode", level=2)
        orf_finder: ORFFinder = PyrodigalFinder(metagenome=True, mask=mask, cpus=jobs)
    else:
        logger.info("Using", f"record features named {cds_feature!r}", level=2)
        orf_finder = CDSFinder(feature=cds_feature, locus_tag=locus_tag)

    unit = "contigs" if len(sequences) > 1 else "contig"

    with logger.progress() as progress:
        task = progress.add_task(
            description="Finding ORFs", total=len(sequences), unit=unit, precision=""
        )

        def callback(record: "SeqRecord", found: int) -> None:
            logger.success("Found", found, "genes in record", repr(record.id), level=2)
            progress.update(task, advance=1)

        return list(orf_finder.find_genes(sequences, progress=callback))


# --- Annotate genes -----------------------------------------------------------


def default_hmms() -> Iterator["HMM"]:
    from ...hmmer import embedded_hmms

    return embedded_hmms()


def custom_hmms(hmm_paths: List[pathlib.Path]) -> Iterable["HMM"]:
    from ...hmmer import HMM

    for path in hmm_paths:
        base = os.path.basename(path)
        file: BinaryIO = zopen(path)
        if base.endswith((".gz", ".lz4", ".xz", ".bz2")):
            base, _ = os.path.splitext(base)
        base, _ = os.path.splitext(base)
        yield HMM(
            id=base,
            version="?",
            url="?",
            path=path,
            size=None,
            relabel_with=r"s/([^\.]*)(\..*)?/\1/",
        )


def filter_domains(
    logger: ConsoleLogger,
    genes: List["Gene"],
    *,
    e_filter: Optional[float] = None,
    p_filter: Optional[float] = None,
) -> List["Gene"]:
    # Filter i-evalue and p-value if required
    if e_filter is not None:
        logger.info("Excluding", "domains with e-value over", e_filter, level=1)
        key = lambda d: d.i_evalue < e_filter
        genes = [
            gene.with_protein(
                gene.protein.with_domains(filter(key, gene.protein.domains))
            )
            for gene in genes
        ]
    if p_filter is not None:
        logger.info("Excluding", "domains with p-value over", p_filter, level=1)
        key = lambda d: d.pvalue < p_filter
        genes = [
            gene.with_protein(
                gene.protein.with_domains(filter(key, gene.protein.domains))
            )
            for gene in genes
        ]
    if p_filter is not None or e_filter is not None:
        count = sum(len(gene.protein.domains) for gene in genes)
        logger.info("Using", "remaining", count, "domains", level=1)
    return genes


def disentangle(gene: "Gene") -> "Gene":
    if len(gene.protein.domains) <= 1:
        return gene

    domains_to_keep = []
    gene_domains = gene.protein.domains.copy()
    while gene_domains:
        # get a domain and collect other overlaping domains
        domain = gene_domains.pop()
        overlaps = [
            other
            for other in gene_domains
            if other.start <= domain.end and domain.start <= other.end
        ]
        # keep only overlapping domain with the best p-value
        if not overlaps or domain.pvalue < min(d.pvalue for d in overlaps):
            domains_to_keep.append(domain)
            for other in overlaps:
                gene_domains.remove(other)

    return gene.with_protein(gene.protein.with_domains(domains_to_keep))


def disentangle_domains(logger, genes: List["Gene"]) -> List["Gene"]:
    logger.info("Disentangling", "overlapping domains in each gene", level=1)
    return list(map(_disentangle, genes))


def annotate_domains(
    logger: ConsoleLogger,
    genes: List["Gene"],
    *,
    hmm_paths: List[pathlib.Path],
    default_hmms: Iterable["HMM"],
    whitelist: Optional[Collection[str]] = None,
    disentangle: bool = False,
    jobs: int = 0,
    bit_cutoffs: Optional[str] = None,
    e_filter: Optional[float] = None,
    p_filter: Optional[float] = None,
) -> List["Gene"]:
    from ...hmmer import PyHMMER

    logger.info("Running", "HMMER domain annotation", level=1)

    # Run all HMMs over ORFs to annotate with protein domains
    hmms = list(custom_hmms(hmm_paths) if hmm_paths else default_hmms)
    total = None if whitelist is None else len(whitelist)

    with logger.progress() as progress:
        task_hmms = progress.add_task(
            description=f"Annotating HMMs",
            unit="file" if len(hmms) == 1 else "files",
            total=len(hmms),
            precision="",
        )
        task_domains = progress.add_task(
            description="", total=total, unit="HMMs", precision=""
        )
        for hmm in progress.track(hmms, task_id=task_hmms, total=len(hmms)):
            progress.update(task_domains, description=f" {hmm.id} v{hmm.version}")
            callback = lambda h, t: progress.update(task_domains, advance=1)
            logger.info(
                "Starting",
                f"annotation with [bold blue]{hmm.id} v{hmm.version}[/]",
                level=2,
            )
            genes = PyHMMER(hmm, jobs, whitelist).run(
                genes, progress=callback, bit_cutoffs=bit_cutoffs
            )
            logger.success(
                "Finished",
                f"annotation with [bold blue]{hmm.id} v{hmm.version}[/]",
                level=2,
            )
        progress.update(task_id=task_domains, visible=False)

    # Count number of annotated domains
    count = sum(len(gene.protein.domains) for gene in genes)
    logger.success("Found", count, "domains across all proteins", level=1)

    # Disentangle if required
    if disentangle:
        genes = _disentangle_domains(logger, genes)

    # Filter i-evalue and p-value if required
    genes = filter_domains(
        logger,
        genes,
        e_filter=e_filter,
        p_filter=p_filter,
    )

    # Sort genes
    logger.info("Sorting", "genes by coordinates", level=2)
    genes.sort(key=operator.attrgetter("source.id", "start", "end"))
    for gene in genes:
        gene.protein.domains.sort(key=operator.attrgetter("start", "end"))

    return genes


# --- Predict probabilities ----------------------------------------------------


def load_model_domains(
    logger: ConsoleLogger,
    classifier: "TypeClassifier",
) -> typing.Set[str]:
    domains = set(classifier.model.attributes_)
    logger.success("Found", len(domains), "selected features", level=2)
    return domains


def predict_probabilities(
    logger: ConsoleLogger,
    genes: List["Gene"],
    *,
    model: Optional[pathlib.Path],
    pad: bool,
    crf_type: Type["ClusterCRF"],
) -> List["Gene"]:
    if model is None:
        logger.info("Loading", "embedded CRF pre-trained model", level=1)
    else:
        logger.info("Loading", "CRF pre-trained model from", repr(str(model)), level=1)
    model = crf_type.trained(model)

    logger.info("Predicting", "cluster probabilitites with the model", level=1)

    with logger.progress() as progress:
        unit = "batches" if len(genes) > 1 else "batch"
        task = progress.add_task("Predicting", unit=unit)

        def progress_callback(i: int, total: int) -> None:
            progress.update(task_id=task, completed=i, total=total)

        return model.predict_probabilities(
            genes,
            pad=pad,
            progress=progress_callback,
        )


def extract_clusters(
    logger: ConsoleLogger,
    genes: List["Gene"],
    *,
    threshold: float,
    postproc: str,
    cds: int,
    edge_distance: int,
    trim: bool,
) -> List["Cluster"]:
    from ...refine import ClusterRefiner

    logger.info("Extracting", "predicted clusters", level=1)
    refiner = ClusterRefiner(
        threshold=threshold,
        criterion=postproc,
        n_cds=cds,
        edge_distance=edge_distance,
        trim=trim,
    )

    with logger.progress() as progress:
        total = len({gene.source.id for gene in genes})
        unit = "contigs" if total > 1 else "contig"
        task = progress.add_task("Segmenting", total=total, unit=unit)
        clusters: List["Cluster"] = []
        gene_groups = itertools.groupby(genes, lambda g: g.source.id)
        for _, gene_group in progress.track(gene_groups, task_id=task, total=total):
            clusters.extend(refiner.iter_clusters(list(gene_group)))

    return clusters


# --- Predict types ------------------------------------------------------------


def load_type_classifier(
    logger: ConsoleLogger,
    *,
    model: Optional[pathlib.Path],
    classifier_type: Type["TypeClassifier"],
) -> "TypeClassifier":
    if model is None:
        logger.info("Loading", "type classifier from embedded model", level=2)
    else:
        logger.info("Loading", "type classifier from", repr(str(model)), level=2)
    return classifier_type.trained(model)


def predict_types(
    logger: ConsoleLogger, clusters: List["Cluster"], *, classifier: "TypeClassifier"
) -> List["Cluster"]:
    logger.info("Predicting", "gene cluster types", level=1)
    with logger.progress() as progress:
        unit = "cluster" if len(clusters) == 1 else "clusters"
        task = progress.add_task("Classifying", unit=unit, total=len(clusters))
        clusters_new = []
        for cluster in progress.track(clusters, task_id=task):
            clusters_new.extend(classifier.predict_types([cluster]))
            if cluster.type:
                name = "/".join(f"[bold blue]{name}[/]" for name in cluster.type.names)
                prob = "/".join(
                    f"[bold purple]{cluster.type_probabilities[name]:.0%}[/]"
                    for name in cluster.type.names
                )
                logger.success(
                    f"Predicted type of [bold blue]{cluster.id}[/] as {name} ({prob} confidence)"
                )
            else:
                ty = max(cluster.type_probabilities, key=cluster.type_probabilities.get)  # type: ignore
                prob = f"[bold purple]{cluster.type_probabilities[ty]:.0%}[/]"
                name = f"[bold blue]{ty}[/]"
                logger.warn(
                    f"Couldn't assign type to [bold blue]{cluster.id}[/] (maybe {name}, {prob} confidence)"
                )
    return clusters_new


# --- Train model --------------------------------------------------------------


def seed_rng(logger: ConsoleLogger, seed: int) -> None:
    logger.info("Seeding", "the random number generator with seed", seed, level=2)
    random.seed(seed)
    numpy.random.seed(seed)


def fit_model(
    logger: ConsoleLogger,
    genes: List["Gene"],
    *,
    feature_type: str,
    c1: float,
    c2: float,
    window_size: int,
    window_step: int,
    shuffle: bool,
    select: Optional[float],
    correction: Optional[str],
    jobs: int = 0,
    crf_type: Type["ClusterCRF"],
) -> "ClusterCRF":
    from ...crf import ClusterCRF

    logger.info("Creating", f"the CRF in [bold blue]{feature_type}[/] mode", level=1)
    logger.info("Using", f"provided hyperparameters (C1={c1}, C2={c2})", level=1)
    if select is not None:
        logger.info(
            "Selecting",
            f"features with Fisher Exact Test (threshold={select})",
            level=1,
        )
    logger.info(
        "Iterating",
        f"over features with a sliding window (W={window_size}, step={window_step})",
        level=1,
    )
    crf = crf_type(
        feature_type,
        algorithm="lbfgs",
        window_size=window_size,
        window_step=window_step,
        c1=c1,
        c2=c2,
    )
    logger.info("Fitting", "the CRF model to the training data")
    crf.fit(
        genes, select=select, shuffle=shuffle, correction_method=correction, cpus=jobs
    )
    return crf



"""A module containing some utilities only relevant to the CLI."""

import contextlib
import io
import os
import typing
import warnings
from types import ModuleType, TracebackType
from typing import Callable, Union, Iterator, TextIO, Type, Optional

import rich.progress
from rich.progress import Column
from rich.text import Text

from .._meta import classproperty, zopen

if typing.TYPE_CHECKING:
    ShowWarning = Callable[
        [Union[Warning, str], Type[Warning], str, int, Optional[TextIO], Optional[str]],
        None,
    ]


class MofNCompleteUnitColumn(rich.progress.MofNCompleteColumn):

    def render(self, task: "Task") -> Text:
        completed = int(task.completed)
        total = int(task.total) if task.total is not None else "?"
        total_width = len(str(total))
        return Text(
            f"{completed:{total_width}d}{self.separator}{total} {task.fields['unit']}",
            style="progress.download",
        )


@contextlib.contextmanager
def patch_showwarnings(new_showwarning: "ShowWarning") -> Iterator[None]:
    """Make a context patching `warnings.showwarning` with the given function."""
    old_showwarning: ShowWarning = warnings.showwarning
    try:
        warnings.showwarning = new_showwarning  # type: ignore
        yield
    finally:
        warnings.showwarning = old_showwarning  # type: ignore


def guess_sequences_format(path: Union[str, "os.PathLike[str]"]) -> Optional[str]:
    """Guess the format of a sequence file located in ``path``.

    Supports the following formats:
        * GenBank
        * FASTA
        * EMBL

    """
    head = ""
    with zopen(path) as file:
        for head in iter(lambda: file.read(256), b""):
            head = head.strip()
            if head:
                break
    if head.startswith(b">"):
        return "fasta"
    elif head.startswith(b"LOCUS"):
        return "genbank"
    elif head.startswith(b"ID "):
        return "embl"
    else:
        return None



import argparse
import os
import operator
from typing import Type, Callable, Iterable, Dict

from rich.console import Console

from . import _parser, _common
from .._log import ConsoleLogger


def configure_parser(
    parser: argparse.ArgumentParser, 
    console: Console,
    program: str,
    version: str,
    *,
    defaults: Dict[str, object],
):
    _parser.configure_common(parser, console, program, version, defaults=defaults)

    _parser.configure_group_input_sequences(parser, short=False, defaults=defaults)
    _parser.configure_group_input_tables(parser, clusters=False, defaults=defaults)
    _parser.configure_group_domain_filter(parser, defaults=defaults)
    _parser.configure_group_cluster_detection(parser, defaults=defaults)
    
    params_output = _parser.configure_group_table_output(parser, defaults=defaults)
    params_output.add_argument(
        "--merge-gbk",
        action="store_true",
        help=(
            "Output a single GenBank file containing every detected cluster "
            "instead of writing one file per cluster."
        ),
    )
    params_output.add_argument(
        "--antismash-sideload",
        action="store_true",
        help=("Write an AntiSMASH v6 sideload JSON file next to the output " "files."),
    )

    parser.set_defaults(run=run)


def run(
    args: argparse.Namespace,
    logger: ConsoleLogger,
    crf_type: Type["ClusterCRF"],
    classifier_type: Type["TypeClassifier"],
    default_hmms: Callable[[], Iterable["HMM"]],
) -> int:
    # attempt to create the output directory, checking it doesn't
    # already contain output files (or raise a warning)
    base, _ = os.path.splitext(os.path.basename(args.genome))
    outputs = [f"{base}.clusters.tsv", f"{base}.features.tsv", f"{base}.genes.tsv"]
    if args.antismash_sideload:
        outputs.append(f"{base}.sideload.json")
    if args.merge_gbk:
        outputs.append(f"{base}.clusters.gbk")
    _common.make_output_directory(logger, args.output_dir, outputs)

    # load features
    genes = list(_common.load_genes(logger, args.genes))
    features = _common.load_features(logger, args.features)

    # load sequences
    sequences = list(
        _common.load_sequences(
            logger,
            args.genome,
            format=args.format,
        )
    )

    # label genes
    genes = _common.annotate_genes(logger, genes, features)
    genes = list(_common.assign_sources(logger, sequences, genes, genome=args.genome))

    # Sort genes
    logger.info("Sorting", "genes by coordinates", level=2)
    genes.sort(key=operator.attrgetter("source.id", "start", "end"))
    for gene in genes:
        gene.protein.domains.sort(key=operator.attrgetter("start", "end"))

    # filter domains by p-value and/or e-value
    genes = _common.filter_domains(
        logger, genes, e_filter=args.e_filter, p_filter=args.p_filter
    )

    # predict probabilities with CRF model
    genes = _common.predict_probabilities(
        logger, genes, model=args.model, pad=args.pad, crf_type=crf_type
    )
    _common.write_genes_table(
        logger, genes, genome=args.genome, output_dir=args.output_dir
    )
    _common.write_feature_table(
        logger, genes, genome=args.genome, output_dir=args.output_dir
    )

    # extract clusters from probability vector
    clusters = _common.extract_clusters(
        logger,
        genes,
        threshold=args.threshold,
        postproc=args.postproc,
        cds=args.cds,
        edge_distance=args.edge_distance,
        trim=args.trim,
    )
    if clusters:
        logger.success("Found", len(clusters), "potential gene clusters", level=1)
    else:
        logger.warn("No gene clusters were found")
        if args.force_tsv:
            _common.write_cluster_table(
                logger,
                clusters,
                genome=args.genome,
                output_dir=args.output_dir,
            )
        return 0

    # predict types for putative clusters
    classifier = _common.load_type_classifier(
        logger,
        model=args.model,
        classifier_type=classifier_type,
    )
    if len(classifier.classes_) > 1:
        clusters = _common.predict_types(logger, clusters, classifier=classifier)

    # write results
    logger.info(
        "Writing", "result files to folder", repr(str(args.output_dir)), level=1
    )
    _common.write_cluster_table(
        logger, clusters, genome=args.genome, output_dir=args.output_dir
    )
    _common.write_clusters(
        logger,
        clusters,
        merge=args.merge_gbk,
        genome=args.genome,
        output_dir=args.output_dir,
    )
    if args.antismash_sideload:
        _common.write_sideload_json(clusters)
    unit = "cluster" if len(clusters) == 1 else "clusters"
    logger.success("Found", len(clusters), "gene", unit, level=0)

    # exit successfully
    return 0



import argparse
import pathlib
from typing import Dict

from rich.console import Console


class HelpExit(SystemExit):
    pass


class ConsoleHelpAction(argparse.Action):

    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help=None,
        deprecated=False,
        console=None,
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )
        self.console = console

    def __call__(self, parser, namespace, values, option_string=None):
        if self.console is None:
            parser.print_help()
        else:
            self.console.print(parser.format_help())
        raise HelpExit(0)


def configure_common(
    parser: argparse.ArgumentParser,
    console: Console,
    program: str,
    version: str,
    *,
    main: bool = False,
    defaults: Dict[str, object],
) -> None:
    parser.add_argument(
        "-h",
        "--help",
        action=ConsoleHelpAction,
        help="Show this help message and exit.",
        console=console,
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"{program} {version}",
        help="Show the program version number and exit.",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        type=int,
        default=0 if main else argparse.SUPPRESS,
        help=(
            "The number of jobs to use for multithreading. Use 0 to use all "
            "available CPUs."
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Increase the console output",
        default=0 if main else argparse.SUPPRESS,
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="count",
        help="Reduce or disable the console output",
        default=0 if main else argparse.SUPPRESS,
    )
    parser.add_argument(
        "--no-color",
        dest="color",
        action="store_false",
        help="Disable the console color",
    )
    parser.add_argument(
        "--no-markup",
        dest="markup",
        action="store_false",
        help="Disable the console markup",
    )


def configure_group_gene_calling(
    parser: argparse.ArgumentParser,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group(
        "Gene Calling",
    )
    group.add_argument(
        "-M",
        "--mask",
        action="store_true",
        help=(
            "Enable unknown region masking to prevent genes from stretching "
            "across ambiguous nucleotides."
        ),
    )
    group.add_argument(
        "--cds-feature",
        type=str,
        help=(
            "Extract genes from annotated records using a feature rather "
            "than running de-novo gene-calling."
        ),
    )
    group.add_argument(
        "--locus-tag",
        type=str,
        default=defaults.get("--locus-tag", "locus_tag"),
        help=(
            "The name of the feature qualifier to use for naming extracted "
            "genes when using the ``--cds-feature`` flag."
        ),
    )
    return group


def configure_group_domain_annotation(
    parser: argparse.ArgumentParser,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group("Domain Annotation")
    group.add_argument(
        "--hmm",
        action="append",
        dest="hmms",
        type=pathlib.Path,
        help=(
            "The path to one or more alternative HMM file to use " "(in HMMER format)."
        ),
    )
    group.add_argument(
        "-e",
        "--e-filter",
        type=float,
        default=defaults.get("--e-filter", None),
        help=(
            "The e-value cutoff for protein domains to be included. This is "
            "not stable across versions, so consider using a p-value filter "
            "instead."
        ),
    )
    group.add_argument(
        "-p",
        "--p-filter",
        type=float,
        default=defaults.get("--p-filter", 1e-9),
        help=("The p-value cutoff for protein domains to be included."),
    )
    group.add_argument(
        "--bit-cutoffs",
        choices=("noise", "gathering", "trusted"),
        default=defaults.get("--bit-cutoffs", None),
        help=("Use HMM-specific bitscore cutoffs to filter domain annotations."),
    )
    # --disentangle creates issues wrt. HMM annotation because it may lead to
    # different domains being selected and applied to a domain depending on
    # whether we annotate with the full HMM or domains from a whitelist only
    # so we disable and hide it by default.
    group.add_argument(
        "--disentangle",
        action="store_true",
        help=argparse.SUPPRESS,
        # help=(
        #     "Disentangle overlapping domains in each gene by keeping only "
        #     "the domains with the lowest E-value over a given position."
        # ),
    )
    return group


def configure_group_domain_filter(
    parser: argparse.ArgumentParser,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group("Domain Annotation")
    group.add_argument(
        "-e",
        "--e-filter",
        type=float,
        default=defaults.get("--e-filter", None),
        help=(
            "The e-value cutoff for protein domains to be included. This is "
            "not stable across versions, so consider using a p-value filter "
            "instead."
        ),
    )
    group.add_argument(
        "-p",
        "--p-filter",
        type=float,
        default=defaults.get("--p-filter", 1e-9),
        help=("The p-value cutoff for protein domains to be included."),
    )
    return group


def configure_group_table_output(
    parser: argparse.ArgumentParser,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group(
        "Output",
    )
    group.add_argument(
        "-o",
        "--output-dir",
        help="The directory in which to write the output files.",
        type=pathlib.Path,
        default=defaults.get("--output-dir", "."),
    )
    group.add_argument(
        "--force-tsv",
        action="store_true",
        help=(
            "Always write TSV output files even when they are empty "
            "(e.g. because no genes or no clusters were found)."
        ),
    )
    return group


def configure_group_cluster_detection(
    parser: argparse.ArgumentParser,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group(
        "Cluster Detection",
    )
    group.add_argument(
        "--model",
        type=pathlib.Path,
        help=(
            "The path to an alternative CRF model to use (obtained with "
            "`gecco train`)."
        ),
    )
    group.add_argument(
        "--no-pad",
        dest="pad",
        action="store_false",
        help=(
            "Disable padding of gene sequences (used to predict gene clusters "
            "in contigs smaller than the CRF window length)."
        ),
    )
    group.add_argument(
        "-c",
        "--cds",
        type=int,
        default=defaults.get("--cds", 3),
        help=(
            "The minimum number of coding sequences a valid cluster must "
            "contain to be retained."
        ),
    )
    group.add_argument(
        "-m",
        "--threshold",
        type=float,
        default=defaults.get("--threshold", 0.8),
        help=("The probability threshold for cluster detection."),
    )
    group.add_argument(
        "--postproc",
        choices={"antismash", "gecco"},
        default=defaults.get("--postproc", "gecco"),
        help=argparse.SUPPRESS,
        # help=(
        #     "The method to use for cluster validation."
        # )
    )
    group.add_argument(
        "-E",
        "--edge-distance",
        default=defaults.get("--edge-distance", 0),
        type=int,
        help=(
            "The minimum number of annotated genes that must separate a "
            "cluster from the edge. Edge clusters will still be included "
            "if they are longer. A lower number will increase the number "
            "of false positives on small contigs."
        ),
    )

    if defaults.get("--trim", True):
        group.add_argument(
            "--no-trim",
            action="store_false",
            dest="trim",
            help=(
                "Disable trimming genes lacking domain annotations from gene "
                "clusters predicted by the internal CRF model."
            ),
        )
    else:
        group.add_argument(
            "--trim",
            action="store_true",
            dest="trim",
            help=(
                "Enable trimming genes lacking domain annotations from gene "
                "clusters predicted by the internal CRF model."
            ),
        )

    return group


def configure_group_training_data(
    parser: argparse.ArgumentParser,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group("Training Data")
    group.add_argument(
        "--no-shuffle",
        action="store_false",
        dest="shuffle",
        help=("Disable shuffling the data before fitting the model."),
    )
    group.add_argument(
        "--seed",
        type=int,
        default=defaults.get("--seed", 42),
        help=(
            "The seed to initialize the random number generator used for "
            "shuffling operations."
        ),
    )
    return group


def configure_group_training_parameters(
    parser: argparse.ArgumentParser,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group("Training Parameters")
    group.add_argument(
        "-W",
        "--window-size",
        type=int,
        default=defaults.get("--window-size", 5),
        help=("The length of the sliding window for CRF predictions."),
    )
    group.add_argument(
        "--window-step",
        type=int,
        default=defaults.get("--window-step", 1),
        help=("The step of the sliding window for CRF predictions."),
    )
    group.add_argument(
        "--c1",
        type=float,
        default=defaults.get("--c1", 0.15),
        help=("The strength of the L1 regularization."),
    )
    group.add_argument(
        "--c2",
        type=float,
        default=defaults.get("--c2", 0.15),
        help=("The strength of the L2 regularization."),
    )
    group.add_argument(
        "--feature-type",
        choices=("protein", "domain"),
        default=defaults.get("--feature-type", "protein"),
        help=(
            "The level at which the features should be extracted and "
            "given to the CRF."
        ),
    )
    group.add_argument(
        "--select",
        type=float,
        default=defaults.get("--select", None),
        help=(
            "The fraction of most significant features to select from "
            "the training data prior to training the CRF."
        ),
    )
    group.add_argument(
        "--correction",
        type=str,
        default=defaults.get("--correction", None),
        help=(
            "The multiple test correction method to use when computing "
            "significance with multiple testing."
        ),
    )
    return group


def configure_group_input_tables(
    parser: argparse.ArgumentParser,
    clusters: bool = True,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group(
        "Input Tables",
    )
    group.add_argument(
        "-f",
        "--features",
        type=pathlib.Path,
        action="append",
        required=True,
        help=("The path to a domain annotation table, used to train the CRF model."),
    )
    group.add_argument(
        "-g",
        "--genes",
        type=pathlib.Path,
        required=True,
        help=(
            "The path to a gene table, containing the coordinates of the genes "
            "inside the training sequence."
        ),
    )
    if clusters:
        group.add_argument(
            "-c",
            "--clusters",
            type=pathlib.Path,
            required=True,
            help=(
                "The path to a cluster annotation table, used to extract the "
                "domain composition for the type classifier."
            ),
        )
    return group


def configure_group_input_sequences(
    parser: argparse.ArgumentParser,
    short: bool = True,
    *,
    defaults: Dict[str, object],
) -> "argparse.ArgumentGroup":
    group = parser.add_argument_group(
        "Input Sequences",
    )
    group.add_argument(
        *["-g"][not short:],
        "--genome",
        required=True,
        help=(
            "A genomic file containing one or more sequences to use as input. "
            "Must be in one of the sequence formats supported by Biopython"
        ),
    )
    group.add_argument(
        *["-f"][not short:],
        "--format",
        help=(
            "The format of the input file, as a Biopython format string. "
            "FASTA, EMBL and GenBank files are recognized automatically "
            "if this is not given."
        ),
    )
    return group


import argparse
import csv
import collections
import itertools
import os
import operator
import pathlib
from typing import List, Optional, Type, Callable, Iterable, Dict

from rich.console import Console

from . import _parser, _common
from .._log import ConsoleLogger


def configure_parser(
    parser: argparse.ArgumentParser, 
    console: Console,
    program: str,
    version: str,
    *,
    defaults: Dict[str, object],
):
    _parser.configure_common(parser, console, program, version, defaults=defaults)

    _parser.configure_group_input_tables(parser, defaults=defaults)  
    _parser.configure_group_domain_filter(parser, defaults=defaults)
    _parser.configure_group_training_data(parser, defaults=defaults)
    _parser.configure_group_training_parameters(parser, defaults=defaults)

    group = parser.add_argument_group(
        "Output",
    )
    group.add_argument(
        "-o",
        "--output-dir",
        help="The directory in which to write the output files.",
        type=pathlib.Path,
        default=defaults.get("--output-dir", pathlib.Path(".")),
    )

    parser.set_defaults(run=run)


def _save_model(
    logger: ConsoleLogger, crf: "ClusterCRF", output_dir: pathlib.Path
) -> None:
    model_out = os.path.join(output_dir, "model.pkl")
    logger.info("Pickling", "the model to", repr(model_out))
    with open(model_out, "wb") as out:
        pickle.dump(crf, out, protocol=4)

    logger.info("Computing", "pickled model checksum", level=2)
    hasher = hashlib.md5()
    with open(model_out, "rb") as out:
        for chunk in iter(lambda: out.read(io.DEFAULT_BUFFER_SIZE), b""):
            hasher.update(chunk)

    logger.info(
        "Writing", "pickled model checksum to", repr(f"{model_out}.md5"), level=2
    )
    with open(f"{model_out}.md5", "w") as out_hash:
        out_hash.write(hasher.hexdigest())


def _save_transitions(
    logger: ConsoleLogger, crf: "ClusterCRF", output_dir: pathlib.Path
) -> None:
    logger.info("Writing", "CRF transitions weights")
    with open(os.path.join(output_dir, "model.trans.tsv"), "w") as f:
        writer = csv.writer(f, dialect="excel-tab")
        writer.writerow(["from", "to", "weight"])
        for labels, weight in crf.model.transition_features_.items():
            writer.writerow([*labels, weight])


def _save_weights(
    logger: ConsoleLogger, crf: "ClusterCRF", output_dir: pathlib.Path
) -> None:
    logger.info("Writing", "state weights")
    with open(os.path.join(output_dir, "model.state.tsv"), "w") as f:
        writer = csv.writer(f, dialect="excel-tab")
        writer.writerow(["attr", "label", "weight"])
        for attrs, weight in crf.model.state_features_.items():
            writer.writerow([*attrs, weight])


def _assign_clusters(
    logger: ConsoleLogger, genes: List["Gene"], clusters: "ClusterTable"
) -> List["Cluster"]:
    from ...model import Cluster, ClusterType

    cluster_types = {}
    cluster_by_seq = collections.defaultdict(list)
    for i in range(len(clusters)):
        seq_id = clusters.sequence_id[i]
        cluster_id = clusters.cluster_id[i]
        cluster_by_seq[seq_id].append(
            (
                clusters.start[i],
                clusters.end[i],
                clusters.cluster_id[i],
            )
        )
        if not "type" in clusters.data.columns:
            cluster_types[cluster_id] = None
        elif clusters.type[i] == "Unknown" or clusters.type[i] is None:
            cluster_types[cluster_id] = ClusterType()
        else:
            cluster_types[cluster_id] = ClusterType(*clusters.type[i].split(";"))

    logger.info("Extracting", "genes belonging to clusters")
    genes_by_cluster = collections.defaultdict(list)
    for seq_id, seq_genes in itertools.groupby(
        genes, key=operator.attrgetter("source.id")
    ):
        for gene in seq_genes:
            for cluster_start, cluster_end, cluster_id in cluster_by_seq[seq_id]:
                if cluster_start <= gene.end and gene.start <= cluster_end:
                    genes_by_cluster[cluster_id].append(gene)

    return [
        Cluster(cluster_id, genes_by_cluster[cluster_id], cluster_types[cluster_id])
        for cluster_id in sorted(filter(None, clusters.cluster_id))
        if genes_by_cluster[cluster_id]
    ]


def _save_domain_compositions(
    logger: ConsoleLogger,
    all_possible: List[str],
    clusters: List["Cluster"],
    *,
    output_dir: pathlib.Path,
) -> None:
    import numpy
    import scipy.sparse

    logger.info("Saving", "training matrix labels for type classifier")
    with open(os.path.join(output_dir, "domains.tsv"), "w") as out:
        out.writelines(f"{domain}\n" for domain in all_possible)
    with open(os.path.join(output_dir, "types.tsv"), "w") as out:
        writer = csv.writer(out, dialect="excel-tab")
        for cluster in clusters:
            types = ";".join(sorted(cluster.type.names))
            writer.writerow([cluster.id, types])

    logger.info("Building", "new domain composition matrix")
    comp = numpy.array([c.domain_composition(all_possible) for c in clusters])

    comp_out = os.path.join(output_dir, "compositions.npz")
    logger.info("Saving", "new domain composition matrix to file", repr(comp_out))
    scipy.sparse.save_npz(comp_out, scipy.sparse.coo_matrix(comp))


def run(
    args: argparse.Namespace,
    logger: ConsoleLogger,
    crf_type: Type["ClusterCRF"],
    classifier_type: Type["TypeClassifier"],
    default_hmms: Callable[[], Iterable["HMM"]],
) -> int:
    # seed RNG
    _common.seed_rng(logger, args.seed)

    # attempt to create the output directory
    outputs = [
        "model.pkl",
        "model.pkl.md5",
        "domains.tsv",
        "types.tsv",
        "compositions.npz",
    ]
    _common.make_output_directory(logger, args.output_dir, outputs)

    # load features
    genes = list(_common.load_genes(logger, args.genes))
    features = _common.load_features(logger, args.features)

    # label genes
    genes = _common.annotate_genes(logger, genes, features)

    # Sort genes
    logger.info("Sorting", "genes by coordinates", level=2)
    genes.sort(key=operator.attrgetter("source.id", "start", "end"))
    for gene in genes:
        gene.protein.domains.sort(key=operator.attrgetter("start", "end"))

    # filter domains by p-value and/or e-value
    genes = _common.filter_domains(
        logger, genes, e_filter=args.e_filter, p_filter=args.p_filter
    )

    # load clusters and label genes inside clusters
    clusters = _common.load_clusters(logger, args.clusters)
    genes = _common.label_genes(logger, genes, clusters)

    # fit CRF
    crf = _common.fit_model(
        logger,
        genes,
        feature_type=args.feature_type,
        c1=args.c1,
        c2=args.c2,
        window_size=args.window_size,
        window_step=args.window_step,
        shuffle=args.shuffle,
        select=args.select,
        correction=args.correction,
        jobs=args.jobs,
        crf_type=crf_type,
    )

    # save model
    logger.info("Saving", f"CRF model to {str(args.output_dir)!r}")
    crf.save(args.output_dir)
    _save_transitions(logger, crf, output_dir=args.output_dir)
    _save_weights(logger, crf, output_dir=args.output_dir)

    # extract the domains names
    logger.info("Finding", "the array of possible protein domains", level=2)
    if crf.significant_features is not None:
        all_possible = sorted(crf.significant_features)
    else:
        all_possible = sorted({d.name for g in genes for d in g.protein.domains})

    # compute domain compositions
    _save_domain_compositions(
        logger,
        all_possible,
        _assign_clusters(logger, genes, clusters),
        output_dir=args.output_dir,
    )
    logger.success("Finished", "training new CRF model", level=0)

    # exit succesfully
    return 0



"""Implementation of the ``gecco annotate`` subcommand."""

import argparse
import os
import typing
from typing import (
    Any,
    BinaryIO,
    Container,
    Collection,
    Dict,
    Iterable,
    Iterator,
    Union,
    Optional,
    List,
    TextIO,
    Mapping,
    Callable,
    Iterable,
    Type,
)

from rich.console import Console

from ..._meta import zopen
from .._log import ConsoleLogger
from . import _parser, _common

if typing.TYPE_CHECKING:
    from Bio.SeqRecord import SeqRecord
    from ...hmmer import HMM
    from ...model import Gene
    from ...orf import ORFFinder


def configure_parser(
    parser: argparse.ArgumentParser, 
    console: Console,
    program: str,
    version: str,
    *,
    defaults: Dict[str, object],
):
    _parser.configure_common(parser, console, program, version, defaults=defaults)

    _parser.configure_group_input_sequences(parser, defaults=defaults)
    _parser.configure_group_gene_calling(parser, defaults=defaults)
    _parser.configure_group_domain_annotation(parser, defaults=defaults)
    _parser.configure_group_table_output(parser, defaults=defaults)

    parser.set_defaults(run=run)


def run(
    args: argparse.Namespace,
    logger: ConsoleLogger,
    crf_type: Type["ClusterCRF"],
    classifier_type: Type["TypeClassifier"],
    default_hmms: Callable[[], Iterable["HMM"]],
) -> int:  # noqa: D102
    # attempt to create the output directory, checking it doesn't
    # already contain output files (or raise a warning)
    base, _ = os.path.splitext(os.path.basename(args.genome))
    outputs = [f"{base}.features.tsv", f"{base}.genes.tsv"]
    _common.make_output_directory(logger, args.output_dir, outputs)

    # load sequences and extract genes
    sequences = list(
        _common.load_sequences(
            logger,
            args.genome,
            format=args.format,
        )
    )
    genes = list(
        _common.extract_genes(
            logger,
            sequences,
            cds_feature=args.cds_feature,
            locus_tag=args.locus_tag,
            mask=args.mask,
            jobs=args.jobs,
        )
    )

    # write gene table
    _common.write_genes_table(
        logger, genes, genome=args.genome, output_dir=args.output_dir
    )
    if genes:
        logger.success("Found", "a total of", len(genes), "genes", level=1)
    else:
        if args.force_tsv:
            _common.write_feature_table([])
        logger.warn("No genes were found")
        return 0

    # annotate genes with protein domains and write results
    genes = _common.annotate_domains(
        logger,
        genes,
        hmm_paths=args.hmms,
        default_hmms=default_hmms(),
        disentangle=args.disentangle,
        jobs=args.jobs,
        bit_cutoffs=args.bit_cutoffs,
        e_filter=args.e_filter,
        p_filter=args.p_filter,
    )
    _common.write_feature_table(
        logger,
        genes,
        genome=args.genome,
        output_dir=args.output_dir,
    )

    # report number of domains found
    ndoms = sum(len(gene.protein.domains) for gene in genes)
    if ndoms:
        logger.success("Found", ndoms, "protein domains", level=0)
    else:
        logger.warn("No protein domains were found")

    # exit successfully
    return 0



import typing
from typing import List, Iterable

import numpy
import sklearn.preprocessing

from ..model import ClusterType

if typing.TYPE_CHECKING:
    from numpy.typing import NDArray


class TypeBinarizer(sklearn.preprocessing.MultiLabelBinarizer):
    """A `MultiLabelBinarizer` working with `ClusterType` instances.
    """

    def __init__(self, classes: List[str], **kwargs: object):
        self.classes_ = classes
        super().__init__(classes=classes, **kwargs)

    def transform(self, y: List[ClusterType]) -> Iterable[Iterable[int]]:
        matrix = numpy.zeros((len(y), len(self.classes_)))
        for i, label in enumerate(y):
            for j, cls in enumerate(self.classes_):
                matrix[i, j] = cls in label.names
        return matrix

    def inverse_transform(self, yt: "NDArray[numpy.bool_]") -> Iterable[ClusterType]:
        classes = []
        for y in yt:
            filtered = (cls for i, cls in enumerate(self.classes_) if y[i])
            classes.append(ClusterType(*filtered))
        return classes



import argparse
import csv
import collections
import itertools
import os
import operator
import pathlib
import random
import typing
from typing import List, Optional, Tuple, Iterable, Callable, Type, Dict

import rich.progress
from rich.console import Console

from . import _parser, _common
from .._log import ConsoleLogger


def configure_parser(
    parser: argparse.ArgumentParser, 
    console: Console,
    program: str,
    version: str,
    *,
    defaults: Dict[str, object],
):
    _parser.configure_common(parser, console, program, version, defaults=defaults)

    _parser.configure_group_input_tables(parser, defaults=defaults)
    _parser.configure_group_domain_filter(parser, defaults=defaults)
    _parser.configure_group_training_data(parser, defaults=defaults)
    _parser.configure_group_training_parameters(parser, defaults=defaults)

    params_cv = parser.add_argument_group("Cross-Validation")
    params_cv.add_argument(
        "--loto",
        action="store_true",
        default=False,
        help=(
            "Use Leave-One-Type-Out (LOTO) cross-validation instead of "
            "K-folds cross-validation."
        ),
    )
    params_cv.add_argument(
        "--splits",
        type=int,
        default=defaults.get("--splits", 10),
        help=("Number of folds for cross-validation (if running K-folds)."),
    )

    params_output = parser.add_argument_group(
        "Output",
    )
    params_output.add_argument(
        "-o",
        "--output",
        type=pathlib.Path,
        default=defaults.get("--output", pathlib.Path("cv.tsv")),
        help=(
            "The name of the output file where the cross-validation table "
            "will be written."
        ),
    )

    parser.set_defaults(run=run)


def _group_genes(
    logger: ConsoleLogger,
    genes: List["Gene"],
    *,
    shuffle: bool,
) -> List[List["Gene"]]:
    logger.info("Grouping", "genes by source sequence")
    groups = itertools.groupby(genes, key=operator.attrgetter("source.id"))
    seqs = [sorted(group, key=operator.attrgetter("start")) for _, group in groups]
    if shuffle:
        logger.info("Shuffling", "training data sequences")
        random.shuffle(seqs)
    return seqs


def _loto_splits(
    logger: ConsoleLogger, seqs: List[List["Gene"]], *, clusters: pathlib.Path
) -> List[Tuple["NDArray[numpy.bool_]", "NDArray[numpy.bool_]"]]:
    from ...crf.cv import LeaveOneGroupOut
    from ...model import ClusterTable, ClusterType

    logger.info("Loading", "the clusters table")
    with zopen(clusters) as in_:
        table = ClusterTable.load(in_)
        index = {row.sequence_id: row.type for row in table}
        if len(index) != len(table):
            raise ValueError("Training data contains several clusters per sequence")

    logger.info("Grouping", "sequences by cluster types")
    groups = []
    for cluster in seqs:
        ty = next((index[g.source.id] for g in cluster if g.source.id in index), None)
        if ty is None:
            seq_id = next(gene.source.id for gene in cluster)
            logger.warn("Failed", f"to find type of cluster in {seq_id!r}")
            ty = ClusterType()
        groups.append(ty.unpack())

    return list(LeaveOneGroupOut().split(seqs, groups=groups))  # type: ignore


def _kfold_splits(
    logger: ConsoleLogger,
    seqs: List[List["Gene"]],
    *,
    splits: int,
) -> List[Tuple["NDArray[numpy.bool_]", "NDArray[numpy.bool_]"]]:
    import sklearn.model_selection

    return list(sklearn.model_selection.KFold(splits).split(seqs))


def _get_train_data(
    train_indices: Iterable[int], seqs: List[List["Gene"]]
) -> List["Gene"]:
    # extract train data
    return [gene for i in train_indices for gene in seqs[i]]


def _get_test_data(
    test_indices: Iterable[int], seqs: List[List["Gene"]]
) -> List["Gene"]:
    # make a clean copy of the test data without gene probabilities
    return [
        gene.with_protein(
            gene.protein.with_domains(
                d.with_probability(None) for d in gene.protein.domains
            )
        )
        for i in test_indices
        for gene in seqs[i]
    ]


def _fit_predict(
    logger,
    train_data: List["Gene"],
    test_data: List["Gene"],
    *,
    crf_type: Type["ClusterCRF"] = None,
    feature_type: str,
    c1: float,
    c2: float,
    window_size: int,
    window_step: int,
    shuffle: bool,
    select: Optional[float],
    correction: Optional[str],
    jobs: int = 0,
) -> List["Gene"]:
    crf = _common.fit_model(
        logger,
        train_data,
        crf_type=crf_type,
        feature_type=feature_type,
        c1=c1,
        c2=c2,
        window_size=window_size,
        window_step=window_step,
        shuffle=shuffle,
        select=select,
        correction=correction,
        jobs=jobs,
    )
    return crf.predict_probabilities(test_data)


def _write_fold(
    logger,
    fold: int,
    truth: List["Genes"],
    predicted: List["Gene"],
    output: pathlib.Path,
    append: bool = False,
) -> None:
    import polars
    from ...model import GeneTable

    frame = (
        GeneTable.from_genes(predicted)
        .data.with_columns(polars.lit(fold).alias("fold"))
        .with_columns(
            GeneTable.from_genes(truth).data.select(
                is_cluster=polars.col("average_p") > 0.5
            )
        )
    )
    with open(output, "ab" if append else "wb") as out:
        frame.write_csv(out, include_header=not append, separator="\t")


def _report_fold(
    logger: ConsoleLogger,
    fold: Optional[int],
    truth: List["Genes"],
    predicted: List["Genes"],
) -> None:
    from sklearn.metrics import average_precision_score, roc_auc_score

    probas = [gene.average_probability for gene in predicted]
    labels = [gene.average_probability > 0.5 for gene in truth]

    aupr = average_precision_score(labels, probas)
    auroc = roc_auc_score(labels, probas)
    if fold:
        logger.info(
            f"Finished training fold {fold} (AUROC={auroc:.3f}, AUPR={aupr:.3f})"
        )
    else:
        logger.info(f"Finished cross validation (AUROC={auroc:.3f}, AUPR={aupr:.3f})")


def run(
    args: argparse.Namespace,
    logger: ConsoleLogger,
    crf_type: Type["ClusterCRF"],
    classifier_type: Type["TypeClassifier"],
    default_hmms: Callable[[], Iterable["HMM"]],
) -> int:
    # seed RNG
    _common.seed_rng(logger, args.seed)

    # load features
    genes = list(_common.load_genes(logger, args.genes))
    features = _common.load_features(logger, args.features)

    # label genes
    genes = _common.annotate_genes(logger, genes, features)

    # load clusters and label genes inside clusters
    clusters = _common.load_clusters(logger, args.clusters)
    genes = _common.label_genes(logger, genes, clusters)

    # group genes
    seqs = _group_genes(logger, genes, shuffle=args.shuffle)
    logger.success("Grouped", "genes into", len(seqs), "sequences")

    # split CV folds
    if args.loto:
        splits = _loto_splits(logger, seqs, clusters=args.clusters)
    else:
        splits = _kfold_splits(logger, seqs, splits=args.splits)

    # run CV
    with logger.progress() as progress:
        unit = "fold" if len(splits) == 1 else "folds"
        task = progress.add_task(
            description="Cross-validating", total=len(splits), unit=unit, precision=""
        )
        logger.info("Performing cross-validation")
        predicted = []
        for i, (train_indices, test_indices) in enumerate(
            progress.track(splits, task_id=task)
        ):
            train_data = _get_train_data(train_indices, seqs)
            test_data = _get_test_data(test_indices, seqs)
            old_genes = _get_train_data(test_indices, seqs)
            new_genes = _fit_predict(
                logger,
                train_data,
                test_data,
                feature_type=args.feature_type,
                c1=args.c1,
                c2=args.c2,
                window_size=args.window_size,
                window_step=args.window_step,
                shuffle=args.shuffle,
                select=args.select,
                correction=args.correction,
                jobs=args.jobs,
                crf_type=crf_type,
            )
            _write_fold(
                logger, i + 1, old_genes, new_genes, output=args.output, append=i > 0
            )
            _report_fold(logger, i + 1, old_genes, new_genes)
            predicted.extend(new_genes)
        _report_fold(logger, None, genes, predicted)



"""Implementation of the ``gecco run`` subcommand."""

import argparse
import os
import typing
import pathlib
from typing import Optional, Type, Iterable, Callable, Dict

from rich.console import Console

from ..._meta import zopen
from .._log import ConsoleLogger
from . import _parser, _common


def configure_parser(
    parser: argparse.ArgumentParser, 
    console: Console,
    program: str,
    version: str,
    *,
    defaults: Dict[str, object],
):
    _parser.configure_common(parser, console, program, version, defaults=defaults)

    _parser.configure_group_input_sequences(parser, defaults=defaults)
    _parser.configure_group_gene_calling(parser, defaults=defaults)
    _parser.configure_group_domain_annotation(parser, defaults=defaults)
    _parser.configure_group_cluster_detection(parser, defaults=defaults)

    params_output = _parser.configure_group_table_output(parser, defaults=defaults)
    params_output.add_argument(
        "--merge-gbk",
        action="store_true",
        help=(
            "Output a single GenBank file containing every detected cluster "
            "instead of writing one file per cluster."
        ),
    )
    params_output.add_argument(
        "--antismash-sideload",
        action="store_true",
        help=("Write an AntiSMASH v6 sideload JSON file next to the output files."),
    )

    parser.set_defaults(run=run)


def run(
    args: argparse.Namespace,
    logger: ConsoleLogger,
    crf_type: Type["ClusterCRF"],
    classifier_type: Type["TypeClassifier"],
    default_hmms: Callable[[], Iterable["HMM"]],
) -> int:
    # attempt to create the output directory, checking it doesn't
    # already contain output files (or raise a warning)
    base, _ = os.path.splitext(os.path.basename(args.genome))
    outputs = [f"{base}.features.tsv", f"{base}.genes.tsv", f"{base}.clusters.tsv"]
    if args.antismash_sideload:
        outputs.append(f"{base}.sideload.json")
    if args.merge_gbk:
        outputs.append(f"{base}.clusters.gbk")
    _common.make_output_directory(logger, args.output_dir, outputs)

    # load sequences and extract genes
    sequences = list(
        _common.load_sequences(
            logger,
            args.genome,
            format=args.format,
        )
    )
    genes = list(
        _common.extract_genes(
            logger,
            sequences,
            cds_feature=args.cds_feature,
            locus_tag=args.locus_tag,
            mask=args.mask,
            jobs=args.jobs,
        )
    )

    # write gene table
    _common.write_genes_table(
        logger, genes, genome=args.genome, output_dir=args.output_dir
    )
    if genes:
        logger.success("Found", "a total of", len(genes), "genes", level=1)
    else:
        if args.force_tsv:
            _common.write_gene_table(
                logger, [], genome=args.genome, output_dir=args.output_dir
            )
            _common.write_feature_table(
                logger, [], genome=args.genome, output_dir=args.output_dir
            )
            _common.write_cluster_table(
                logger, [], genome=args.genome, output_dir=args.output_dir
            )
        logger.warn("No genes were found")
        return 0

    # load type classifier (to access compositional data)
    classifier = _common.load_type_classifier(
        logger,
        model=args.model,
        classifier_type=classifier_type,
    )

    # generate whitelist from internal feature list of model
    whitelist = _common.load_model_domains(
        logger,
        classifier,
    )

    # annotate genes with protein domains
    genes = _common.annotate_domains(
        logger,
        genes,
        hmm_paths=args.hmms,
        default_hmms=default_hmms(),
        whitelist=whitelist,
        disentangle=args.disentangle,
        jobs=args.jobs,
        bit_cutoffs=args.bit_cutoffs,
        e_filter=args.e_filter,
        p_filter=args.p_filter,
    )

    # predict gene cluster probabilities
    genes = _common.predict_probabilities(
        logger,
        genes,
        model=args.model,
        pad=args.pad,
        crf_type=crf_type,
    )

    # write gene and feature tables
    _common.write_genes_table(
        logger,
        genes,
        genome=args.genome,
        output_dir=args.output_dir,
    )
    _common.write_feature_table(
        logger,
        genes,
        genome=args.genome,
        output_dir=args.output_dir,
    )

    # extract clusters from probability vector
    clusters = _common.extract_clusters(
        logger,
        genes,
        threshold=args.threshold,
        postproc=args.postproc,
        cds=args.cds,
        edge_distance=args.edge_distance,
        trim=args.trim,
    )
    if clusters:
        logger.success("Found", len(clusters), "potential gene clusters", level=1)
    else:
        logger.warn("No gene clusters were found")
        if args.force_tsv:
            _common.write_cluster_table(
                logger,
                clusters,
                genome=args.genome,
                output_dir=args.output_dir,
            )
        return 0

    # predict types for putative clusters if the type classifier has more
    # than one class to predict from
    if len(classifier.classes_) > 1:
        clusters = _common.predict_types(logger, clusters, classifier=classifier)

    # write results
    logger.info(
        "Writing", "result files to folder", repr(str(args.output_dir)), level=1
    )
    _common.write_cluster_table(
        logger, clusters, genome=args.genome, output_dir=args.output_dir
    )
    _common.write_clusters(
        logger,
        clusters,
        merge=args.merge_gbk,
        genome=args.genome,
        output_dir=args.output_dir,
    )
    if args.antismash_sideload:
        _common.write_sideload_json(clusters)
    unit = "cluster" if len(clusters) == 1 else "clusters"
    logger.success("Found", len(clusters), "gene", unit, level=0)

    # exit successfully
    return 0



import abc
import datetime
import os
import socket
from typing import Any, List, Type, Optional, TextIO

import rich.progress
from rich.console import Console
from rich.progress import Progress, DownloadColumn

from .. import __name__ as PROGRAM
from ._utils import MofNCompleteUnitColumn


class ConsoleLogger(abc.ABC):
    console: Console

    def __init__(
        self,
        console: Console,
        *,
        quiet: int = 0,
        verbose: int = 0,
        hostname: Optional[str] = None,
        pid: Optional[int] = None,
        program: str = PROGRAM,
    ):
        self.console = console
        self.verbose = verbose
        self.quiet = quiet
        self._hostname = socket.gethostname() if hostname is None else hostname
        self._pid = os.getpid() if pid is None else pid
        self._program = program

    @abc.abstractmethod
    def error(self, message: str, *args: Any, level: int = 0) -> None:
        pass

    @abc.abstractmethod
    def info(self, verb: str, *args: Any, level: int = 1) -> None:
        pass

    @abc.abstractmethod
    def success(self, verb: str, *args: Any, level: int = 1) -> None:
        pass

    @abc.abstractmethod
    def warn(self, verb: str, *args: Any, level: int = 0) -> None:
        pass

    @abc.abstractmethod
    def progress(self, *args, **kwargs) -> Progress:
        pass


class FullConsoleLogger(ConsoleLogger):

    def _logprefix(self) -> List[str]:
        return [
            f"[bold dim cyan]{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}[/]",
            f"[bold dim purple]{self._hostname}[/]",
            f"[bold dim]{self._program}[[default dim]{self._pid}[/]][/]",
        ]

    def error(self, message: str, *args: Any, level: int = 0) -> None:
        if self.quiet <= 2 and level <= self.verbose:
            self.console.print(
                *self._logprefix(),
                "[bold red]FAIL[/]",
                message,
                *args,
            )

    def info(self, verb: str, *args: Any, level: int = 1) -> None:
        if self.quiet == 0 and level <= self.verbose:
            self.console.print(
                *self._logprefix(),
                f"[bold blue]INFO[/]",
                verb,
                *args,
            )

    def success(self, verb: str, *args: Any, level: int = 1) -> None:
        if self.quiet == 0 and level <= self.verbose:
            self.console.print(
                *self._logprefix(),
                f"[bold green]  OK[/]",
                verb,
                *args,
            )

    def warn(self, verb: str, *args: Any, level: int = 0) -> None:
        if self.quiet <= 1 and level <= self.verbose:
            self.console.print(*self._logprefix(), "[bold yellow]WARN[/]", verb, *args)

    def progress(self, download: bool = False) -> Progress:
        column = DownloadColumn() if download else MofNCompleteUnitColumn()
        return Progress(
            *self._logprefix(),
            f"[bold purple]WORK[/]",
            rich.progress.TextColumn("[progress.description]{task.description}"),
            rich.progress.BarColumn(),
            rich.progress.TaskProgressColumn(),
            rich.progress.TimeElapsedColumn(),
            column,
            console=self.console,
            transient=True,
        )


class ConciseConsoleLogger(ConsoleLogger):

    def error(self, message: str, *args: Any, level: int = 0) -> None:
        if self.quiet <= 2 and level <= self.verbose:
            self.console.print(
                "[bold red]x[/]",
                message,
                *args,
            )

    def info(self, verb: str, *args: Any, level: int = 1) -> None:
        if self.quiet == 0 and level <= self.verbose:
            self.console.print(
                "[bold blue]i[/]",
                verb,
                *args,
            )

    def success(self, verb: str, *args: Any, level: int = 1) -> None:
        if self.quiet == 0 and level <= self.verbose:
            self.console.print(
                "[green]:heavy_check_mark:[/]",
                verb,
                *args,
            )

    def warn(self, verb: str, *args: Any, level: int = 0) -> None:
        if self.quiet <= 1 and level <= self.verbose:
            self.console.print(
                "[bold yellow]![/]",
                verb,
                *args,
            )

    def progress(self, download: bool = False) -> Progress:
        column = DownloadColumn() if download else MofNCompleteUnitColumn()
        return Progress(
            rich.progress.SpinnerColumn(finished_text="[green]:heavy_check_mark:[/]"),
            rich.progress.TextColumn("[progress.description]{task.description:<18}"),
            rich.progress.BarColumn(),
            rich.progress.TaskProgressColumn(),
            rich.progress.TimeElapsedColumn(),
            column,
            console=self.console,
        )


def make_logger(
    console: Console,
    verbose: int,
    **kwargs,
):
    if verbose >= 2:
        return FullConsoleLogger(
            console=console,
            verbose=verbose,
            **kwargs,
        )
    else:
        return ConciseConsoleLogger(
            console=console,
            verbose=verbose,
            **kwargs,
        )


def showwarnings(
    logger: ConsoleLogger,
    message: str,
    category: Type[Warning],
    filename: str,
    lineno: int,
    file: Optional[TextIO] = None,
    line: Optional[str] = None,
    verbose: int = 0,
    quiet: int = 0,
) -> None:
    for line in filter(str.strip, str(message).splitlines()):
        logger.warn(line.strip())



"""Algorithm to smooth contiguous gene cluster predictions into single regions.
"""

import copy
import collections.abc
import itertools
import functools
import operator
import typing
from typing import List, Mapping, Optional, Tuple, Iterator

import numpy
from Bio.SeqRecord import SeqRecord

from .model import Cluster, Domain, Gene, Protein, Strand


__all__ = ["BIO_PFAMS", "GeneGrouper", "ClusterRefiner"]


# fmt: off
# `set` of `str`: A set of domains from Pfam considered 'biosynthetic' by AntiSMASH.
BIO_PFAMS = frozenset({
    "PF00109", "PF02801", "PF08659", "PF00378", "PF08541", "PF08545",
    "PF02803", "PF00108", "PF02706", "PF03364", "PF08990", "PF00501",
    "PF00668", "PF08415", "PF00975", "PF03061", "PF00432", "PF00494",
    "PF03936", "PF01397", "PF00432", "PF04275", "PF00348", "PF02401",
    "PF04551", "PF00368", "PF00534", "PF00535", "PF02922", "PF01041",
    "PF00128", "PF00908", "PF02719", "PF04321", "PF01943", "PF02806",
    "PF02350", "PF02397", "PF04932", "PF01075", "PF00953", "PF01050",
    "PF03033", "PF01501", "PF05159", "PF04101", "PF02563", "PF08437",
    "PF02585", "PF01721", "PF02052", "PF02674", "PF03515", "PF04369",
    "PF08109", "PF08129", "PF09221", "PF09683", "PF10439", "PF11420",
    "PF11632", "PF11758", "PF12173", "PF04738", "PF04737", "PF04604",
    "PF05147", "PF08109", "PF08129", "PF08130", "PF00155", "PF00202",
    "PF00702", "PF06339", "PF04183", "PF10331", "PF03756", "PF00106",
    "PF01370", "PF00107", "PF08240", "PF00441", "PF02770", "PF02771",
    "PF08028", "PF01408", "PF02894", "PF00984", "PF00725", "PF03720",
    "PF03721", "PF07993", "PF02737", "PF00903", "PF00037", "PF04055",
    "PF00171", "PF00067", "PF01266", "PF01118", "PF02668", "PF00248",
    "PF01494", "PF01593", "PF03992", "PF00355", "PF01243", "PF00384",
    "PF01488", "PF00857", "PF04879", "PF08241", "PF08242", "PF00698",
    "PF00483", "PF00561", "PF00583", "PF01636", "PF01039", "PF00288",
    "PF00289", "PF02786", "PF01757", "PF02785", "PF02409", "PF01553",
    "PF02348", "PF00891", "PF01596", "PF04820", "PF02522", "PF08484",
    "PF08421",
})



class GeneGrouper:
    """A callable to group genes under or over a probability threshold.

    Use with a list of genes in combination with `itertools.groupby`.
    """

    def __init__(self, threshold: float):  # noqa: D102, D107
        self.in_cluster = False
        self.threshold = threshold

    def __call__(self, gene: Gene) -> bool:  # noqa: D102
        if gene.average_probability is not None:
            self.in_cluster = gene.average_probability > self.threshold
        return self.in_cluster



class ClusterRefiner:
    """A post-processor to extract contiguous clusters from CRF predictions.
    """

    def __init__(
        self,
        *,
        threshold: float = 0.8,
        criterion: str = "gecco",
        n_cds: int = 5,
        n_biopfams: int = 5,
        average_threshold: float = 0.6,
        edge_distance: int = 0,
        trim: bool = True,
    ) -> None:
        """Create a new `ClusterRefiner` instance.

        Arguments:
            threshold (`float`): The probability threshold to use to consider
                a protein to be part of a gene cluster.
            criterion (`str`): The criterion to use when checking for cluster
                validity. 
            n_cds (`int`): The minimum number of genes a gene cluster must
                contain to be considered valid. If ``criterion`` is ``gecco``,
                then this is the minimum number of **annotated** CDS.
            n_biopfams (`int`): The minimum number of biosynthetic Pfam
                domains a gene cluster must contain to be considered valid
                (*only when the criterion is* ``antismash``).
            average_threshold (`int`): The average probability threshold to
                use to consider a gene cluster valid (*only when the 
                criterion is* ``antismash``).
            edge_distance (`int`): The minimum distance from the edge the 
                gene cluster must be located (it may start at an edge, but must 
                span for longer than ``edge_distance``), in number of annotated 
                genes (*only when the criterion is* ``gecco``).
            trim (`bool`): If set to `True` (the default), raw segments 
                predicted by a `~gecco.crf.ClusterCRF` will be post-processed
                to exclude genes on cluster edges which have no domain 
                annotation. Set to `False` to retain all genes as predicted
                by the CRF.

        """
        self.threshold = threshold
        self.criterion = criterion
        self.n_cds = n_cds
        self.n_biopfams = n_biopfams
        self.average_threshold = average_threshold
        self.edge_distance = edge_distance
        self.trim = trim

    def iter_clusters(self, genes: List[Gene]) -> Iterator[Cluster]:
        """Find all clusters in a table of CRF predictions.

        Arguments:
            genes (`list` of `~gecco.model.Gene`): A list of genes with
                probability annotations estimated by `~gecco.crf.ClusterCRF`.

        Yields:
            `gecco.model.Cluster`: Valid clusters found in the input with
            respect to the postprocessing criterion given at initialisation.

        """
        for seq, cluster in self._iter_clusters(genes):
            if self.trim:
                cluster = self._trim_cluster(cluster)
            if self._validate_cluster(seq, cluster):
                yield cluster
        #
        # unfiltered_clusters = map(self._trim_cluster, self._iter_clusters(genes))
        # return filter(self._validate_cluster, unfiltered_clusters)

    def _validate_cluster(self, seq: List[Gene], cluster: Cluster) -> bool:
        """Check a cluster validity depending on the postprocessing criterion.
        """
        if self.criterion == "gecco":
            # check that the number of cluster genes that are outside of
            # the edge region is above the number of required CDS
            annotated = [g for g in cluster.genes if g.protein.domains]
            cds_crit = len(annotated) >= self.n_cds
            # extract IDs of annotated genes that are too close to the edge
            if self.edge_distance > 0:
                annotated_ids = [g.id for g in seq if g.protein.domains]
                edge_genes = set(annotated_ids[:self.edge_distance]).union(annotated_ids[-self.edge_distance:])
            else:
                edge_genes = set()
            # NOTE (@althonos): it is needed for compatibility with the post-processed results
            #                   that we filter on any number of cluster genes, but it would be
            #                   better to filter on the number of annotated cluster genes instead.
            edge_crit = len(set(g.id for g in cluster.genes).difference(edge_genes)) >= self.n_cds
            return cds_crit and edge_crit
        elif self.criterion == "antismash":
            domains = {d.name for gene in cluster.genes for d in gene.protein.domains}
            p_crit = numpy.mean([typing.cast(float, g.average_probability) for g in cluster.genes]) >= self.average_threshold
            bio_crit = len(domains & BIO_PFAMS) >= self.n_biopfams
            cds_crit = len(cluster.genes) >= self.n_cds
            return p_crit and bio_crit and cds_crit
        else:
            raise ValueError(f"Unknown cluster filtering criterion: {self.criterion}")

    def _trim_cluster(self, cluster: Cluster) -> Cluster:
        """Remove unannotated proteins from the cluster edges.
        """
        genes = cluster.genes.copy()
        while genes and not genes[0].protein.domains:
            genes.pop(0)
        while genes and not genes[-1].protein.domains:
            genes.pop()
        return Cluster(
            cluster.id, 
            genes, 
            cluster.type, 
            cluster.type_probabilities
        )

    def _iter_clusters(
        self,
        genes: List[Gene],
    ) -> Iterator[Tuple[List[Gene], Cluster]]:
        """Iterate over contiguous cluster segments from a list of genes.
        """
        grouper = GeneGrouper(self.threshold)
        key = operator.attrgetter("source.id")

        # iterate over the genes grouped by sequence ids
        for seq_id, sequence in itertools.groupby(sorted(genes, key=key), key=key):
            # sort genes within the same sequence by coordinates
            seqsort = sorted(sequence, key=operator.attrgetter("start", "end"))
            # group contiguous genes if they are over or under the threshold
            groups = itertools.groupby(seqsort, key=grouper)
            # filter out regions that are not identified to be clusters
            clusters = (genes for in_cluster, genes in groups if in_cluster)
            for i, cluster in enumerate(clusters):
                yield seqsort, Cluster(id=f"{seq_id}_cluster_{i+1}", genes=list(cluster))



# PYTHON_ARGCOMPLETE_OK

import argparse
import functools
import signal
import typing
from typing import Optional, List, TextIO, Type, Callable, Iterable, Dict

from rich.console import Console

try:
    from rich_argparse import ArgumentDefaultsRichHelpFormatter
except ImportError:
    ArgumentDefaultsRichHelpFormatter = None

try:
    import argcomplete
except ImportError:
    argcomplete = None

from ... import __version__, __package__ as _program
from .._utils import patch_showwarnings
from .._log import showwarnings, make_logger, ConsoleLogger
from ._common import default_hmms
from ._parser import ConsoleHelpAction
from . import annotate, run, predict, train, cv, convert

if typing.TYPE_CHECKING:
    from ...hmmer import HMM
    from ...crf import ClusterCRF


def _formatter_class(console):
    if ArgumentDefaultsRichHelpFormatter is None:
        return argparse.ArgumentDefaultsHelpFormatter
    else:
        return functools.partial(ArgumentDefaultsRichHelpFormatter, console=console)


def configure_parser(
    parser: argparse.ArgumentParser,
    console: Console,
    program: str,
    version: str,
    *,
    defaults: Dict[str, object],
) -> argparse.ArgumentParser:
    _parser.configure_common(parser, console, program, version, main=True, defaults=defaults)

    commands = parser.add_subparsers(required=True, metavar="COMMAND")
    annotate.configure_parser(
        commands.add_parser(
            "annotate",
            formatter_class=_formatter_class(console),
            help="Annotate protein features of one or several contigs.",
            add_help=False,
        ),
        console,
        program,
        version,
        defaults=defaults,
    )
    run.configure_parser(
        commands.add_parser(
            "run",
            formatter_class=_formatter_class(console),
            help="Predict gene clusters from one or several contigs.",
            add_help=False,
        ),
        console,
        program,
        version,
        defaults=defaults,
    )
    predict.configure_parser(
        commands.add_parser(
            "predict",
            formatter_class=_formatter_class(console),
            help="Predict gene clusters on contigs that have been annotated.",
            add_help=False,
        ),
        console,
        program,
        version,
        defaults=defaults,
    )
    train.configure_parser(
        commands.add_parser(
            "train",
            formatter_class=_formatter_class(console),
            help="Train a new CRF model on pre-generated tables.",
            add_help=False,
        ),
        console,
        program,
        version,
        defaults=defaults,
    )
    cv.configure_parser(
        commands.add_parser(
            "cv",
            formatter_class=_formatter_class(console),
            help="Train and evaluate a model using cross-validation.",
            add_help=False,
        ),
        console,
        program,
        version,
        defaults=defaults,
    )
    convert.configure_parser(
        commands.add_parser(
            "convert",
            formatter_class=_formatter_class(console),
            help="Convert output files to a different format.",
            add_help=False,
        ),
        console,
        program,
        version,
        defaults=defaults,
    )

    return parser


def main(
    argv: Optional[List[str]] = None,
    console: Optional[Console] = None,
    *,
    program: str = _program,
    version: str = __version__,
    default_hmms: Callable[[], Iterable["HMM"]] = default_hmms,
    crf_type: Optional[Type["ClusterCRF"]] = None,
    classifier_type: Optional[Type["TypeClassifier"]] = None,
    defaults: Optional[Dict[str, object]] = None,
) -> int:
    if defaults is None:
        defaults = {}

    parser = configure_parser(
        argparse.ArgumentParser(
            prog=program,
            formatter_class=_formatter_class(console),
            add_help=False,
        ),
        console,
        program,
        version,
        defaults=defaults,
    )
    if argcomplete is not None:
        argcomplete.autocomplete(parser)

    try:
        args = parser.parse_args(argv)
    except _parser.HelpExit:
        return 0

    if crf_type is None:
        from ...crf import ClusterCRF

        crf_type = ClusterCRF

    if classifier_type is None:
        from ...types import TypeClassifier

        classifier_type = TypeClassifier

    if console is None:
        console = Console(
            legacy_windows=not args.markup,
            no_color=not args.color,
            quiet=args.quiet,
            safe_box=not args.markup,
        )

    logger = make_logger(
        console,
        quiet=args.quiet,
        verbose=args.verbose,
        program=program,
    )

    with patch_showwarnings(
        functools.partial(showwarnings, logger, verbose=args.verbose, quiet=args.quiet)
    ):
        try:
            return args.run(
                args,
                logger,
                crf_type=crf_type,
                classifier_type=classifier_type,
                default_hmms=default_hmms,
            )
        except KeyboardInterrupt:
            logger.error("Interrupted")
            return -signal.SIGINT
        except OSError as e:
            logger.error(f"{e.strerror}: {e.filename!r}")
            return e.errno
        except RuntimeError as e:
            logger.error(f"{e.strerror}: {e.filename!r}")
            return 1
        except Exception as err:
            console.print_exception()
            if hasattr(err, "code"):
                return err.code
            elif hasattr(err, "errno"):
                return err.errno
            return 1
        else:
            return 0



import argparse
import csv
import glob
import os
import pathlib
import re
from typing import Type, Callable, Iterable, Dict

import rich.progress
from rich.console import Console

try:
    from rich_argparse import ArgumentDefaultsRichHelpFormatter as DefaultFormatter
except ImportError:
    from argparse import ArgumentDefaultsHelpFormatter as DefaultFormatter

from .._log import ConsoleLogger
from . import _parser


def configure_parser(
    parser: argparse.ArgumentParser, 
    console: Console,
    program: str,
    version: str,
    *,
    defaults: Dict[str, object],
):
    _parser.configure_common(parser, console, program, version, defaults=defaults)

    commands = parser.add_subparsers(required=True, metavar="INPUT")

    gbk_parser = commands.add_parser(
        "gbk",
        formatter_class=DefaultFormatter,
        help="Convert the GenBank records to a different format.",
    )
    gbk_parser.set_defaults(input="gbk")
    gbk_parser.add_argument(
        "-i",
        "--input-dir",
        type=pathlib.Path,
        required=True,
        help=("The path to the input directory containing files to convert."),
    )
    gbk_parser.add_argument(
        "-o",
        "--output-dir",
        type=pathlib.Path,
        help=("The path to the directory where to write converted files."),
        default=None,
    )
    gbk_parser.add_argument(
        "-f",
        "--format",
        required=True,
        choices=("bigslice", "fna", "faa"),
        help=("The output format to write."),
    )

    clusters_parser = commands.add_parser(
        "clusters",
        formatter_class=DefaultFormatter,
        help="Convert the clusters table to a different format.",
    )
    clusters_parser.set_defaults(input="clusters")
    clusters_parser.add_argument(
        "-i",
        "--input-dir",
        type=pathlib.Path,
        required=True,
        help=("The path to the input directory containing files to convert."),
    )
    clusters_parser.add_argument(
        "-o",
        "--output-dir",
        type=pathlib.Path,
        help=("The path to the directory where to write converted files."),
        default=None,
    )
    clusters_parser.add_argument(
        "-f",
        "--format",
        required=True,
        choices=("gff",),
        help=("The output format to write."),
    )

    parser.set_defaults(run=run)

    return parser


# --- GenBank to BiG-SLICE GenBank ---------------------------------------------


def _convert_gbk_bigslice(
    logger: ConsoleLogger,
    input_dir: pathlib.Path,
    output_dir: pathlib.Path,
    progress: rich.progress.Progress,
) -> None:
    import Bio.SeqIO
    from Bio.SeqFeature import SeqFeature, FeatureLocation
    from ...model import ClusterTable

    # collect `*.clusters.tsv` files
    cluster_files = glob.glob(os.path.join(input_dir, "*.clusters.tsv"))
    unit = "table" if len(cluster_files) == 1 else "tables"
    task = progress.add_task(
        "Loading clusters", total=len(cluster_files), unit=unit, precision=""
    )
    # load the original coordinates from the `*.clusters.tsv` files
    coordinates = {}
    types = {}
    for cluster_file in progress.track(cluster_files, task_id=task):
        table = ClusterTable.load(cluster_file)
        for i, cluster_id in enumerate(table.cluster_id):
            coordinates[cluster_id] = (table.start[i], table.end[i])
            types[cluster_id] = table.type[i] or "Unknown"

    # collect `*_clusters_{N}.gbk` files
    gbk_files = list(input_dir.glob("*_cluster_*.gbk"))
    unit = "file" if len(gbk_files) == 1 else "files"
    task = progress.add_task(
        "Converting", total=len(gbk_files), unit=unit, precision=""
    )
    done = 0
    # rewrite GenBank files
    for gbk_file in progress.track(gbk_files, task_id=task, total=len(gbk_files)):
        # load record and ensure it comes from GECCO
        record = Bio.SeqIO.read(gbk_file, "genbank")
        if "GECCO-Data" not in record.annotations.get("structured_comment", {}):
            logger.warn(f"GenBank file {gbk_file!r} was not obtained by GECCO")
            continue
        # mark as AntiSMASH v5
        record.annotations["structured_comment"]["antiSMASH-Data"] = {
            "Version": "5.X",
            "Orig. start": coordinates[record.id][0],
            "Orig. end": coordinates[record.id][1],
        }
        # using a /subregion feature will make BiG-SLiCE think the
        # cluster is coming from MIBiG, and the predicted type will be
        # handled correctly
        subregion_feature = SeqFeature(
            FeatureLocation(0, len(record)), type="subregion"
        )
        subregion_feature.qualifiers["contig_edge"] = ["False"]
        subregion_feature.qualifiers["aStool"] = ["mibig"]
        subregion_feature.qualifiers["label"] = [types[record.id]]
        record.features.append(subregion_feature)
        # rename the {id}_cluster_{N}.gbk file to {id}.region{N}.gbk
        contig_id, cluster_n = re.search(r"^(.*)_cluster_(\d+).gbk", gbk_file.name).groups()  # type: ignore
        new_name = output_dir.joinpath(
            "{}.region{:03}.gbk".format(contig_id, int(cluster_n))
        )
        logger.info(f"Rewriting {str(gbk_file)!r} to {str(new_name)!r}")
        Bio.SeqIO.write(record, new_name, "genbank")
        done += 1
    logger.success("Converted", f"{done} GenBank {unit} to BiG-SLiCE format", level=0)


# --- GenBank to nucleotide FASTA ----------------------------------------------


def _convert_gbk_fna(
    logger: ConsoleLogger,
    input_dir: pathlib.Path,
    output_dir: pathlib.Path,
    progress: rich.progress.Progress,
) -> None:
    import Bio.SeqIO
    from Bio.SeqFeature import SeqFeature, FeatureLocation

    # collect `*_clusters_{N}.gbk` files
    gbk_files = list(input_dir.glob("*_cluster_*.gbk"))
    unit = "file" if len(gbk_files) == 1 else "files"
    task = progress.add_task(
        "Converting", total=len(gbk_files), unit=unit, precision=""
    )
    done = 0
    # rewrite GenBank files
    for gbk_file in progress.track(gbk_files, task_id=task, total=len(gbk_files)):
        # load record and ensure it comes from GECCO
        record = Bio.SeqIO.read(gbk_file, "genbank")
        if "GECCO-Data" not in record.annotations.get("structured_comment", {}):
            logger.warn(f"GenBank file {gbk_file!r} was not obtained by GECCO")
            continue
        # convert to nucleotide FASTA
        new_name = output_dir.joinpath(gbk_file.with_suffix(".fna").name)
        logger.info(f"Converting {str(gbk_file)!r} to FASTA file {str(new_name)!r}")
        Bio.SeqIO.write(record, new_name, "fasta")
        done += 1
    logger.success(
        "Converted", f"{done} GenBank {unit} to nucleotide FASTA format", level=0
    )


# --- GenBank to protein FASTA -------------------------------------------------


def _convert_gbk_faa(
    logger: ConsoleLogger,
    input_dir: pathlib.Path,
    output_dir: pathlib.Path,
    progress: rich.progress.Progress,
) -> None:
    import Bio.SeqIO
    from Bio.Seq import Seq
    from Bio.SeqRecord import SeqRecord
    from Bio.SeqFeature import SeqFeature, FeatureLocation

    # collect `*_clusters_{N}.gbk` files
    gbk_files = list(input_dir.glob("*_cluster_*.gbk"))
    unit = "file" if len(gbk_files) == 1 else "files"
    task = progress.add_task(
        "Converting", total=len(gbk_files), unit=unit, precision=""
    )
    done = 0
    # rewrite GenBank files
    for gbk_file in progress.track(gbk_files, task_id=task, total=len(gbk_files)):
        # load record and ensure it comes from GECCO
        record = Bio.SeqIO.read(gbk_file, "genbank")
        if "GECCO-Data" not in record.annotations.get("structured_comment", {}):
            logger.warn(f"GenBank file {gbk_file!r} was not obtained by GECCO")
            continue
        # extract proteins
        proteins = []
        for feat in filter(lambda f: f.type == "CDS", record.features):
            if "locus_tag" not in feat.qualifiers:
                continue
            seq = Seq(feat.qualifiers["translation"][0])
            prot = SeqRecord(id=feat.qualifiers["locus_tag"][0], seq=seq)
            proteins.append(prot)
        # write proteins to FASTA
        new_name = output_dir.joinpath(gbk_file.with_suffix(".faa").name)
        logger.info(f"Converting {str(gbk_file)!r} proteins to FASTA file {str(new_name)!r}")
        Bio.SeqIO.write(proteins, new_name, "fasta")
        done += 1
    logger.success(
        "Converted", f"{done} GenBank {unit} to protein FASTA format", level=0
    )


# --- Clusters TSV to GFF ------------------------------------------------------


def _convert_clusters_gff(
    logger: ConsoleLogger,
    input_dir: pathlib.Path,
    output_dir: pathlib.Path,
    progress: rich.progress.Progress,
) -> None:
    import Bio.SeqIO
    from ...model import ClusterTable

    # collect `*_clusters_{N}.gbk` files
    tsv_files = list(input_dir.glob("*.clusters.tsv"))
    unit = "file" if len(tsv_files) == 1 else "files"
    task = progress.add_task(
        "Converting", total=len(tsv_files), unit=unit, precision=""
    )
    done = 0

    # rewrite GenBank files
    for tsv_file in progress.track(tsv_files, task_id=task, total=len(tsv_files)):
        table = ClusterTable.load(tsv_file)
        gff_file = output_dir.joinpath(tsv_file.with_suffix(".gff").name)
        with gff_file.open("w") as dst:
            writer = csv.writer(dst, dialect="excel-tab")
            writer.writerow(["##gff-version 3"])
            for row in table.data.rows(named=True):
                # load the GenBank files of the cluster to extract the GECCO version
                gbk_file = os.path.join(input_dir, f"{row['cluster_id']}.gbk")
                cluster = Bio.SeqIO.read(gbk_file, "genbank")
                annotations = cluster.annotations["structured_comment"]["GECCO-Data"]
                # make sure to have a BGC type to write down
                bgc_types = ["Unknown"] if not row["type"] else row["type"].split(";")
                # extract type probabilities
                type_probas = []
                for key, value in row.items():
                    if key.endswith("_probability"):
                        ty = key.split("_")[0].capitalize()
                        if ty == "Nrp":
                            ty = "NRP"
                        type_probas.append(f"Type{ty}={value}")
                # write the GFF row
                writer.writerow(
                    [
                        row["sequence_id"],
                        annotations["version"],
                        "BGC",
                        str(row["start"]),
                        str(row["end"]),
                        str(row["average_p"]),
                        ".",
                        ".",
                        ";".join(
                            [
                                f"ID={row['cluster_id']}",
                                f"Name={ '/'.join(sorted(bgc_types)) } cluster",
                                f"Type={ ','.join(sorted(bgc_types)) }",
                                f"ProbabilityAverage={row['average_p']}",
                                f"ProbabilityMax={row['max_p']}",
                                *type_probas,
                                f"Genes={row['proteins'].count(';') + 1}",
                                f"Domains={row['domains'].count(';') + 1}",
                            ]
                        ),
                    ]
                )
        done += 1
    logger.success("Converted", f"{done} TSV {unit} to GFF format", level=0)


# --- Run command --------------------------------------------------------------


def run(
    args: argparse.Namespace,
    logger: ConsoleLogger,
    crf_type: Type["ClusterCRF"],
    classifier_type: Type["TypeClassifier"],
    default_hmms: Callable[[], Iterable["HMM"]],
) -> int:

    input_dir = args.input_dir
    output_dir = args.output_dir if args.output_dir is not None else input_dir
    output_dir.mkdir(parents=True, exist_ok=True)

    with logger.progress() as progress:
        if args.input == "gbk":
            if args.format == "bigslice":
                _convert_gbk_bigslice(logger, input_dir, output_dir, progress)
            elif args.format == "fna":
                _convert_gbk_fna(logger, input_dir, output_dir, progress)
            elif args.format == "faa":
                _convert_gbk_faa(logger, input_dir, output_dir, progress)
            else:
                raise NotImplementedError
        elif args.input == "clusters":
            if args.format == "gff":
                _convert_clusters_gff(logger, input_dir, output_dir, progress)
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

    # exit successfully
    return 0
